/*
 Copyright (C) 2007 Christian Dywan <christian@twotoasts.de>

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "webView.h"

// #define MIDORI_WEBKIT_SUCHEN
// #define MIDORI_WEBKIT_VERGROESSERN
// #define MIDORI_WEBKIT_MARKIERUNG // bug #16144

#ifdef MIDORI_WEBKIT_ANTWORT
WebKitNavigationResponse on_webView_navigation_requested(GtkWidget* webView
 , WebKitFrame* frame, WebKitNetworkRequest* networkRequest)
{
    //g_print("nr: %s\n", webkit_network_request_get_url(networkRequest));
    /*WEBKIT_NAVIGATION_RESPONSE_IGNORE,
    WEBKIT_NAVIGATION_RESPONSE_DOWNLOAD*/
    return WEBKIT_NAVIGATION_RESPONSE_ACCEPT;
}
#endif

void on_webView_location_changed(GtkWidget* webView, const gchar* uri
, CBrowser* browser)
{
    // TODO: Ask webkit wether it knows the protocol for "unknown protcol"
    // TODO: This isn't the place for uri scheme handling
    // TODO: Does this check suffice to indicate an uri?
    // FIXME: sProtocol is leaking, fix that beautifully
    gchar* protocol = strtok(g_strdup(uri), ":");
    if(spawn_protocol_command(protocol, uri))
    {
        //webView_stop(webView); // TODO: This hack would crash webkit
        return;
    }
    gchar* newUri = g_strdup(uri ? uri : "");
    xbel_bookmark_set_href(browser->sessionItem, newUri);
    if(webView == get_nth_webView(-1, browser))
    {
        gtk_entry_set_text(GTK_ENTRY(browser->location), newUri);
        gtk_label_set_text(GTK_LABEL(browser->webView_name), newUri);
        update_status_message(NULL, browser);
        update_gui_state(browser);
    }
}

void on_webView_title_changed(GtkWidget* webView, const gchar* title
 , const gchar* uri, CBrowser* browser)
{
    // TODO: We emulate location_changed here for now
    // Shouldn't we have separated title_changed and location_changed signals?
    on_webView_location_changed(webView, uri, browser);
    gchar* newTitle;
    if(title)
        newTitle = g_strdup(title);
    else
    {
        const gchar* uri = xbel_bookmark_get_href(browser->sessionItem);
        newTitle = g_strdup(uri);
    }
    xbel_item_set_title(browser->sessionItem, newTitle);
    gtk_label_set_text(GTK_LABEL(browser->webView_name), newTitle);
    sokoke_widget_set_tooltip_text(gtk_widget_get_parent(
     gtk_widget_get_parent(browser->webView_name)), newTitle);
    gtk_label_set_text(GTK_LABEL(gtk_bin_get_child(GTK_BIN(
     browser->webView_menu))), newTitle);
    if(webView == get_nth_webView(-1, browser))
    {
     gchar* windowTitle = g_strconcat(newTitle, " - ", PACKAGE_NAME, NULL);
     gtk_window_set_title(GTK_WINDOW(browser->window), windowTitle);
     g_free(windowTitle);
    }
}

/*
void on_webFrame_title_changed(WebKitFrame* frame, const gchar* sTitle
 , const gchar* sUri, CBrowser* browser)
{
    UNIMPLEMENTED
}
*/

#ifdef MIDORI_WEBKIT
void on_webView_icon_changed(GtkWidget* webView, WebKitFrame* widget
 , CBrowser* browser)
{
    // TODO: Implement icon updates; currently this isn't ever called anyway
    const gchar* icon = NULL;
    UNIMPLEMENTED
    if(icon)
    {
     gtk_label_set_text(GTK_LABEL(browser->webView_name), "icon");
     gtk_label_set_text(GTK_LABEL(gtk_bin_get_child(GTK_BIN(
      browser->webView_menu))), "icon");
     if(webView == get_nth_webView(-1, browser))
     {
      gchar* windowTitle = g_strconcat("icon", " - ", PACKAGE_NAME, NULL);
      gtk_window_set_title(GTK_WINDOW(browser->window), windowTitle);
      g_free(windowTitle);
     }
    }
}

void on_webView_load_started(GtkWidget* webView, WebKitFrame* widget
 , CBrowser* browser)
{
    browser->loadedPercent = 0;
    browser->loadedBytes = 0;
    browser->loadedBytesMax = 0;
    update_favicon(browser);
    if(webView == get_nth_webView(-1, browser)) update_gui_state(browser);
    update_statusbar_text(browser);
    gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(browser->progress), 0.1);
    gtk_widget_show(browser->progress);
}
#endif

void on_webView_load_changed(GtkWidget* webView, gint progress, CBrowser* browser)
{
    browser->loadedBytes = progress;
    browser->loadedBytes = 100;
    if(webView == get_nth_webView(-1, browser)) update_gui_state(browser);
}

#ifdef MIDORI_WEBKIT
void on_webView_load_finished(GtkWidget* webView, WebKitFrame* widget
 , CBrowser* browser)
{
    browser->loadedPercent = -1;
    update_favicon(browser);
    if(webView == get_nth_webView(-1, browser)) update_gui_state(browser);
    update_statusbar_text(browser);
    gtk_widget_hide(browser->progress);
}
#endif

/*
void on_frame_cleared(WebKitFrame* frame, CBrowser* browser)
{
    UNIMPLEMENTED
}

void on_frame_load_done(WebKitFrame* frame, gboolean bSuccess
 , CBrowser* browser)
{
    UNIMPLEMENTED
}
*/

void on_webView_status_message(GtkWidget* webView, const gchar* text, CBrowser* browser)
{
    update_status_message(text, browser);
}

void on_webView_selection_changed(GtkWidget* webView, CBrowser* browser)
{
    UNIMPLEMENTED
}

gboolean on_webView_console_message(GtkWidget* webView
 , const gchar* message, gint line, const gchar* sourceId, CBrowser* browser)
{
    return FALSE;
}

typedef enum
{
 SCRIPT_DIALOG_ALERT,
 SCRIPT_DIALOG_CONFIRM,
 SCRIPT_DIALOG_PROMPT
} SCRIPT_DIALOG;

static gboolean webView_script_dialog(GtkWidget* webView, gpointer frame
 , const gchar* message, SCRIPT_DIALOG type, const gchar* defaultValue
 , gchar** value, CBrowser* browser)
{
    GtkMessageType messageType = GTK_MESSAGE_OTHER;
    GtkButtonsType buttons = GTK_BUTTONS_NONE;
    gint defaultResponse = 0;
    switch(type)
    {
     case SCRIPT_DIALOG_ALERT:
      messageType = GTK_MESSAGE_WARNING;
      buttons = GTK_BUTTONS_CLOSE;
      defaultResponse = GTK_RESPONSE_CLOSE; break;
     case SCRIPT_DIALOG_CONFIRM:
      messageType = GTK_MESSAGE_QUESTION;
      buttons = GTK_BUTTONS_YES_NO;
      defaultResponse = GTK_RESPONSE_YES; break;
     case SCRIPT_DIALOG_PROMPT:
      messageType = GTK_MESSAGE_QUESTION;
      buttons = GTK_BUTTONS_OK_CANCEL;
      defaultResponse = GTK_RESPONSE_OK;
    }

    GtkWidget* dialog = gtk_message_dialog_new(GTK_WINDOW(browser->window)
     , GTK_DIALOG_DESTROY_WITH_PARENT, messageType, buttons, "%s", message);
    // FIXME: use the frame's uri
    gchar* title = g_strconcat("JavaScript - ", webView_get_uri(browser->webView), NULL);
    gtk_window_set_title(GTK_WINDOW(dialog), title); g_free(title);

    GtkWidget* entry = NULL;
    if(type == SCRIPT_DIALOG_PROMPT)
    {
     entry = gtk_entry_new();
     gtk_entry_set_text(GTK_ENTRY(entry), defaultValue);
     gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), entry);
     gtk_entry_set_activates_default(GTK_ENTRY(entry), TRUE);
     gtk_widget_show(entry);
    }

    gtk_dialog_set_default_response(GTK_DIALOG(dialog), defaultResponse);
    gtk_widget_set_sensitive(webView, FALSE);
    gint response = sokoke_dialog_run_modeless(GTK_DIALOG(dialog));
    gtk_widget_set_sensitive(webView, TRUE);

    gboolean returnValue = FALSE;
    switch(response)
    {
    case GTK_RESPONSE_YES:
        returnValue = TRUE;
        break;
    case GTK_RESPONSE_OK:
        returnValue = TRUE;
        *value = g_strdup(gtk_entry_get_text(GTK_ENTRY(entry)));
    }
    gtk_widget_destroy(dialog);
    return returnValue;
}

gboolean on_webView_script_alert(GtkWidget* webView, gpointer frame
 , const gchar* message, CBrowser* browser)
{
    webView_script_dialog(webView, frame, message
     , SCRIPT_DIALOG_ALERT, NULL, NULL, browser);
    return TRUE;
}

gboolean on_webView_script_confirm(GtkWidget* webView, gpointer frame
 , const gchar* message, gboolean* confirmed, CBrowser* browser)
{
    *confirmed = webView_script_dialog(webView, frame, message
     , SCRIPT_DIALOG_CONFIRM, NULL, NULL, browser);
    return TRUE;
}

gboolean on_webView_script_prompt(GtkWidget* webView, gpointer frame
 , const gchar* message, const gchar* defaultValue, gchar** value, CBrowser* browser)
{
    if(!webView_script_dialog(webView, frame, message
     , SCRIPT_DIALOG_PROMPT, defaultValue, value, browser))
     *value = NULL;
    return TRUE;
}

/*
void on_webView_request_auth_prompt(GtkWidget* webView, CBrowser* browser)
{
    // gchar* relam; gchar* url;
    // gchar* out_username; gchar* out_password; gboolean out_ok_pressed;
    GtkDialog* dialog;
    gchar* title; gchar* msg;
    GtkWidget* hbox;
    GtkSizeGroup* sizegroup; GtkWidget* align; GtkWidget* label;
    GtkWidget* user; GtkWidget* pass;
    gint response;

    msg = g_strdup_printf("Authentication required for:\n\n%s", realm);
    dialog = GTK_DIALOG(gtk_message_dialog_new(GTK_WINDOW(browser->window)
     , GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_OTHER, GTK_BUTTONS_OK_CANCEL
     , msg)); g_free(msg);
    title = g_strconcat("Authentication - ", url, NULL);
    gtk_window_set_title(GTK_WINDOW(dialog), title); g_free(title);
    #if GTK_CHECK_VERSION(2, 10, 0)
    GtkWidget* image = gtk_image_new_from_icon_name(
     GTK_STOCK_DIALOG_AUTHENTICATION, GTK_ICON_SIZE_DIALOG);
    gtk_widget_show(image);
    gtk_message_dialog_set_image(GTK_MESSAGE_DIALOG(dialog), image);
    #endif
    sizegroup = gtk_size_group_new(GTK_SIZE_GROUP_HORIZONTAL);
    hbox = gtk_hbox_new(FALSE, 6);
    label = gtk_label_new("Username");
    align = gtk_alignment_new(0, 0.5, 0, 0);
    gtk_container_add(GTK_CONTAINER(align), label);
    gtk_size_group_add_widget(sizegroup, align);
    gtk_box_pack_start(GTK_BOX(hbox), align, FALSE, FALSE, 0);
    user = gtk_entry_new();
    gtk_box_pack_start(GTK_BOX(hbox), user, FALSE, FALSE, 0);
    gtk_container_add(GTK_CONTAINER(dialog->vbox), hbox);
    gtk_widget_show_all(hbox);
    hbox = gtk_hbox_new(FALSE, 6);
    label = gtk_label_new("Password");
    align = gtk_alignment_new(0, 0.5, 0, 0);
    gtk_container_add(GTK_CONTAINER(align), label);
    gtk_size_group_add_widget(sizegroup, align);
    gtk_box_pack_start(GTK_BOX(hbox), align, FALSE, FALSE, 0);
    pass = gtk_entry_new();
    gtk_entry_set_visibility(GTK_ENTRY(pass), FALSE);
    gtk_box_pack_start(GTK_BOX(hbox), pass, FALSE, FALSE, 0);
    gtk_container_add(GTK_CONTAINER(dialog->vbox), hbox);
    gtk_widget_show_all(hbox);
    gtk_dialog_set_default_response(dialog, GTK_RESPONSE_CANCEL);

    gtk_widget_set_sensitive(webView, FALSE);
    response = sokoke_dialog_run_modeless(dialog);
    gtk_widget_set_sensitive(webView, TRUE);

    if(response == GTK_RESPONSE_OK)
    {
     out_username = g_strdup(gtk_entry_get_text(GTK_ENTRY(user)));
     out_password = g_strdup(gtk_entry_get_text(GTK_ENTRY(pass)));
     out_ok_pressed = TRUE;
    }
    else
     out_ok_pressed = FALSE;
    gtk_widget_destroy(GTK_WIDGET(dialog));
}
*/

/*
gboolean on_webView_set_cookie(GtkWidget* webView, CBrowser* browser)
{
    // Whole cookie
    // const gchar* cookie;
    // Cookie name
    // const gchar* name;
    // Cookie value
    // const gchar* value;
    // Cookie comment
    // const gchar* comment;
    // Cookie domain
    // const gchar* domain;
    // Cookie TTL
    // const gchar* maxAge;
    // Cookie path
    // const gchar* path;
    // Is cookie secure?
    // gboolean secure;
    // Cookie-specification version.
    // gint version;

    return TRUE;
}
*/

void on_webView_link_hover(GtkWidget* webView, const gchar* tooltip
 , const gchar* uri, CBrowser* browser)
{
    update_status_message(uri, browser);
    sokoke_widget_set_visible(browser->icon_link_script
     , uri && g_str_has_prefix(uri, "javascript:"));
    g_free(browser->elementUri); browser->elementUri = g_strdup(uri);
}

/*
GtkWidget* on_webView_window_open(GtkWidget* webView, const gchar* sUri
 , CBrowser* browser)
{
    // A window is created
    // TODO: Respect config->iNewPages
    // TODO: Find out if this comes from a script or a click
    // TODO: Block scripted popups, return NULL and show status icon
    CBrowser* newBrowser = browser_new(config->openPopupsInTabs ? browser : NULL);
    return newBrowser->webView;
}
*/

/*
void on_webView_window_features(GtkWidget* webView, CBrowser* browser)
{
    browser->hasMenubar = TRUE;
    browser->hasToolbar = properties->toolbarsVisible;
    browser->hasLocation = properties->toolbarsVisible;
    browser->hasStatusbar = properties->statusBarVisible;
    //browser->hasScrollbars = properties->scrollbarsVisible;
    browser->isResizable = properties->isResizable;
    update_visibility(browser, TRUE);
}
*/

void webView_popup(GtkWidget* webView, GdkEventButton* event, CBrowser* browser)
{
    gboolean isLink = browser->elementUri != NULL; // Did we right-click a link?
    gboolean haveLinks = FALSE; // TODO: Are several links selected?
    gboolean isImage = FALSE; // TODO: Did we right-click an image?
    gboolean isEditable = webView_can_paste_clipboard(webView); //...
    gchar* selection = webView_get_selected_text(get_nth_webView(-1, browser));

    update_edit_items(browser);

    // Download manager available?
    const gchar* downloadManager = g_datalist_get_data(&config->protocols_commands, "download");
    gboolean canDownload = downloadManager && *downloadManager;

    action_set_visible("LinkTabNew", isLink, browser);
    action_set_visible("LinkTabCurrent", isLink, browser);
    action_set_visible("LinkWindowNew", isLink, browser);

    action_set_visible("LinkSaveAs", isLink, browser);
    action_set_visible("LinkSaveWith", isLink && canDownload, browser);
    action_set_visible("LinkCopy", isLink, browser);
    action_set_visible("LinkBookmarkNew", isLink, browser);

    action_set_visible("SelectionLinksNewTabs", haveLinks && selection && *selection, browser);
    action_set_visible("SelectionTextTabNew", haveLinks && selection && *selection, browser);
    action_set_visible("SelectionTextTabCurrent", haveLinks && selection && *selection, browser);
    action_set_visible("SelectionTextWindowNew", haveLinks && selection && *selection, browser);

    action_set_visible("ImageViewTabNew", isImage, browser);
    action_set_visible("ImageViewTabCurrent", isImage, browser);
    action_set_visible("ImageSaveAs", isImage, browser);
    action_set_visible("ImageSaveWith", isImage && canDownload, browser);
    action_set_visible("ImageCopy", isImage, browser);

    action_set_visible("Copy_", (selection && *selection) || isEditable, browser);
    action_set_visible("SelectionSearch", selection && *selection, browser);
    action_set_visible("SelectionSearchWith", selection && *selection, browser);
    action_set_visible("SelectionSourceView", selection && *selection, browser);

    action_set_visible("SourceView", !selection, browser);

    if(isEditable)
        sokoke_widget_popup(webView, GTK_MENU(browser->popup_editable), event);
    else if(isLink || isImage || (selection && *selection))
        sokoke_widget_popup(webView, GTK_MENU(browser->popup_element), event);
    else
        sokoke_widget_popup(webView, GTK_MENU(browser->popup_webView), event);
}

gboolean on_webView_button_release(GtkWidget* widget, GdkEventButton* event
 , CBrowser* browser)
{
    GdkModifierType state = (GdkModifierType)0;
    gint x, y;
    gdk_window_get_pointer(NULL, &x, &y, &state);
    switch(event->button)
    {
    case 1:
        if(!browser->elementUri) return FALSE;
        if(state & GDK_SHIFT_MASK)
        {
            // Open link in new window
            CBrowser* curBrowser = browser_new(NULL);
            webView_load_from_uri(curBrowser->webView, browser->elementUri);
            return TRUE;
        }
        else if(state & GDK_MOD1_MASK)
        {
            // Open link in new tab
            CBrowser* curBrowser = browser_new(browser);
            webView_load_from_uri(curBrowser->webView, browser->elementUri);
            return TRUE;
        }
        break;
    case 2:
        if(state & GDK_CONTROL_MASK)
        {
            webView_set_text_size(get_nth_webView(-1, browser), 1);
            return TRUE;
        }
        else
        {
            if(!browser->elementUri) return FALSE;
            // Open link in new tab
            CBrowser* curBrowser = browser_new(browser);
            webView_load_from_uri(curBrowser->webView, browser->elementUri);
            return TRUE;
        }
        break;
    case 3:
        webView_popup(get_nth_webView(-1, browser), event, browser);
        return TRUE;
    }
    return FALSE;
}

void on_webView_popup(GtkWidget* webView, CBrowser* browser)
{
    webView_popup(get_nth_webView(-1, browser), NULL, browser);
}

gboolean on_webView_scroll(GtkWidget* webView, GdkEventScroll* event
 , CBrowser* browser)
{
    GdkModifierType state = (GdkModifierType)0;
    gint x, y;
    gdk_window_get_pointer(NULL, &x, &y, &state);
    if(state & GDK_CONTROL_MASK)
    {
        const gfloat size = webView_get_text_size(get_nth_webView(-1, browser));
        if(event->direction == GDK_SCROLL_DOWN)
            webView_set_text_size(get_nth_webView(-1, browser), size + 0.1);
        else if(event->direction == GDK_SCROLL_UP)
            webView_set_text_size(get_nth_webView(-1, browser), size - 0.1);
        return TRUE;
    }
    else
        return FALSE;
}

gboolean on_webView_leave(GtkWidget* webView, GdkEventCrossing* event, CBrowser* browser)
{
    update_status_message(NULL, browser);
    return TRUE;
}

void on_webView_destroy(GtkWidget* widget, CBrowser* browser)
{
    // Update browser list, free memory and possibly quit
    GList* tmp = g_list_find(browsers, browser);
    browsers = g_list_delete_link(browsers, tmp);
    g_free(browser->elementUri);
    g_free(browser->statusMessage);
    if(!g_list_length(browsers))
    {
        g_object_unref(browser->actiongroup);
        g_object_unref(browser->popup_webView);
        g_object_unref(browser->popup_element);
        g_object_unref(browser->popup_editable);
        gtk_main_quit();
    }
}

// webView actions begin here

GtkWidget* webView_new(GtkWidget** scrolled)
{
    *scrolled = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(*scrolled)
     , GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
    GTK_WIDGET_SET_FLAGS(*scrolled, GTK_CAN_FOCUS);
    #ifdef MIDORI_WEBKIT
    GtkWidget* webView = webkit_page_new();
    gtk_container_add(GTK_CONTAINER(*scrolled), webView);
    #else
    GtkWidget* webView = gtk_event_box_new();
    gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(*scrolled), webView);
    #endif
    return webView;
}

#define webView_IS_SENSITIVE(rv) if(!GTK_WIDGET_SENSITIVE(webView)) return rv

void webView_load_from_uri(GtkWidget* webView, const gchar* uri)
{
    webView_IS_SENSITIVE();
    #ifdef MIDORI_WEBKIT
    webkit_page_open(WEBKIT_PAGE(webView), (gchar*)uri);
    // We need to check the browser first
    // No browser means this is a panel
    CBrowser* browser = get_browser_from_webView(webView);
    if(browser)
    {
        xbel_bookmark_set_href(browser->sessionItem, uri);
        xbel_item_set_title(browser->sessionItem, "");
    }
    #endif
}

void webView_load_from_string(GtkWidget* webView, const gchar* string)
{
    webView_IS_SENSITIVE();
    #ifdef MIDORI_WEBKIT
    webkit_page_load_string(WEBKIT_PAGE(webView), string, "text/plain", "UTF-8", "");
    #endif
}

gchar* webView_copy_source(GtkWidget* webView)
{
    g_return_val_if_fail(GTK_IS_WIDGET(webView), NULL);
    #ifdef MIDORI_WEBKIT_GIBTSNICHT
    WebKitFrame* frame = webkit_page_get_main_frame(WEBKIT_PAGE(webView));
    return webkit_frame_copy_source(frame);
    #else
    return NULL;
    #endif
}

void webView_go_back(GtkWidget* webView)
{
    webView_IS_SENSITIVE();
    #ifdef MIDORI_WEBKIT
    webkit_page_go_backward(WEBKIT_PAGE(webView));
    #endif
}

void webView_go_forward(GtkWidget* webView)
{
    webView_IS_SENSITIVE();
    #ifdef MIDORI_WEBKIT
    webkit_page_go_forward(WEBKIT_PAGE(webView));
    #endif
}

void webView_refresh(GtkWidget* webView, gboolean bSkipCache)
{
    webView_IS_SENSITIVE();
    #ifdef MIDORI_WEBKIT
    webkit_page_reload(WEBKIT_PAGE(webView));
    #endif
}

void webView_stop(GtkWidget* webView)
{
    webView_IS_SENSITIVE();
    #ifdef MIDORI_WEBKIT
    webkit_page_stop_loading(WEBKIT_PAGE(webView));
    #endif
}

gboolean webView_can_go_back(GtkWidget* webView)
{
    g_return_val_if_fail(GTK_IS_WIDGET(webView), FALSE);
    #ifdef MIDORI_WEBKIT
    return webkit_page_can_go_backward(WEBKIT_PAGE(webView));
    #else
    return FALSE;
    #endif
}

gboolean webView_can_go_forward(GtkWidget* webView)
{
    g_return_val_if_fail(GTK_IS_WIDGET(webView), FALSE);
    #ifdef MIDORI_WEBKIT
    return webkit_page_can_go_forward(WEBKIT_PAGE(webView));
    #else
    return FALSE;
    #endif
}

G_CONST_RETURN gchar* webView_get_uri(GtkWidget* webView)
{
    g_return_val_if_fail(GTK_IS_WIDGET(webView), NULL);
    #ifdef MIDORI_WEBKIT
    WebKitFrame* frame = webkit_page_get_main_frame(WEBKIT_PAGE(webView));
    return webkit_frame_get_location(frame);
    #else
    return NULL;
    #endif
}

G_CONST_RETURN gchar* webView_get_title(GtkWidget* webView)
{
    g_return_val_if_fail(GTK_IS_WIDGET(webView), NULL);
    #ifdef MIDORI_WEBKIT
    WebKitFrame* frame = webkit_page_get_main_frame(WEBKIT_PAGE(webView));
    return webkit_frame_get_title(frame);
    #else
    return NULL;
    #endif
}


gboolean webView_find(GtkWidget* webView, const gchar* text
 , gboolean forward, gboolean caseSensitive, gboolean wrap)
{
    g_return_val_if_fail(GTK_IS_WIDGET(webView), FALSE);
    webView_IS_SENSITIVE(FALSE);
    #ifdef MIDORI_WEBKIT_SUCHEN
    return webkit_page_search_for(WEBKIT_PAGE(webView)
     , text, forward, caseSensitive, TRUE);
    #else
    return FALSE;
    #endif
}

gboolean webView_can_increase_text_size(GtkWidget* webView)
{
    g_return_val_if_fail(GTK_IS_WIDGET(webView), FALSE);
    webView_IS_SENSITIVE(FALSE);
    #ifdef MIDORI_WEBKIT_VERGROESSERN
    return webkit_page_can_increase_text_size(WEBKIT_PAGE(webView));
    #else
    return FALSE;
    #endif
}

gboolean webView_can_decrease_text_size(GtkWidget* webView)
{
    g_return_val_if_fail(GTK_IS_WIDGET(webView), FALSE);
    webView_IS_SENSITIVE(FALSE);
    #ifdef MIDORI_WEBKIT_VERGROESSERN
    return webkit_page_can_decrease_text_size(WEBKIT_PAGE(webView));
    #else
    return FALSE;
    #endif
}

void webView_set_text_size(GtkWidget* webView, gfloat size)
{
    webView_IS_SENSITIVE();
    #ifdef MIDORI_WEBKIT_VERGROESSERN
    webkit_page_set_text_size_multiplier(WEBKIT_PAGE(webView), size);
    #endif
}

gfloat webView_get_text_size(GtkWidget* webView)
{
    g_return_val_if_fail(GTK_IS_WIDGET(webView), 1);
    #ifdef MIDORI_WEBKIT_VERGROESSERN
    return webkit_page_get_text_size_multiplier(WEBKIT_PAGE(webView));
    #else
    return FALSE;
    #endif
}

gboolean webView_can_cut_clipboard(GtkWidget* webView)
{
    g_return_val_if_fail(GTK_IS_WIDGET(webView), FALSE);
    #ifdef MIDORI_WEBKIT_MARKIERUNG
    return webkit_page_can_cut_clipboard(WEBKIT_PAGE(webView));
    #else
    return FALSE;
    #endif
}

gboolean webView_can_copy_clipboard(GtkWidget* webView)
{
    g_return_val_if_fail(GTK_IS_WIDGET(webView), FALSE);
    #ifdef MIDORI_WEBKIT_MARKIERUNG
    return webkit_page_can_copy_clipboard(WEBKIT_PAGE(webView));
    #else
    return FALSE;
    #endif
}

gboolean webView_can_paste_clipboard(GtkWidget* webView)
{
    g_return_val_if_fail(GTK_IS_WIDGET(webView), FALSE);
    #ifdef MIDORI_WEBKIT_MARKIERUNG
    return webkit_page_can_paste_clipboard(WEBKIT_PAGE(webView));
    #else
    return FALSE;
    #endif
}

void webView_cut_clipboard(GtkWidget* webView)
{
    g_return_if_fail(GTK_IS_WIDGET(webView));
    #ifdef MIDORI_WEBKIT_MARKIERUNG
    webkit_page_cut_clipboard(WEBKIT_PAGE(webView));
    #endif
}

void webView_copy_clipboard(GtkWidget* webView)
{
    g_return_if_fail(GTK_IS_WIDGET(webView));
    #ifdef MIDORI_WEBKIT_MARKIERUNG
    webkit_page_copy_clipboard(WEBKIT_PAGE(webView));
    #endif
}

void webView_paste_clipboard(GtkWidget* webView)
{
    g_return_if_fail(GTK_IS_WIDGET(webView));
    #ifdef MIDORI_WEBKIT_MARKIERUNG
    webkit_page_paste_clipboard(WEBKIT_PAGE(webView));
    #endif
}

void webView_delete_selection(GtkWidget* webView)
{
    g_return_if_fail(GTK_IS_WIDGET(webView));
    #ifdef MIDORI_WEBKIT_MARKIERUNG
    webkit_page_delete_selection(WEBKIT_PAGE(webView));
    #endif
}

void webView_select_region(GtkWidget* webView, gint start, gint end)
{
    g_return_if_fail(GTK_IS_WIDGET(webView));
    #ifdef MIDORI_WEBKIT_MARKIERUNG
    webkit_page_select_region(WEBKIT_PAGE(webView), start, end);
    #endif
}

gchar* webView_get_selected_text(GtkWidget* webView)
{
    g_return_val_if_fail(GTK_IS_WIDGET(webView), NULL);
    #ifdef MIDORI_WEBKIT_MARKIERUNG
    return webkit_page_get_selected_text(WEBKIT_PAGE(webView));
    #else
    return NULL;
    #endif
}

gchar* webView_get_chars(GtkWidget* webView, gint start, gint end)
{
    g_return_val_if_fail(GTK_IS_WIDGET(webView), NULL);
    #ifdef MIDORI_WEBKIT_MARKIERUNG
    return webkit_page_get_chars(WEBKIT_PAGE(webView), start, end);
    #else
    return NULL;
    #endif
}

gboolean webView_get_selection_bounds(GtkWidget* webView, gint* start, gint* end)
{
    g_return_val_if_fail(GTK_IS_WIDGET(webView), FALSE);
    #ifdef MIDORI_WEBKIT_MARKIERUNG
    return webkit_page_get_selection_bounds(WEBKIT_PAGE(webView), start, end);
    #else
    return FALSE;
    #endif
}

void webView_close(GtkWidget* webView, CBrowser* browser)
{
    browser = get_browser_from_webView(webView);
    // TODO: Delete old items if > 10 items
    const gchar* uri = xbel_bookmark_get_href(browser->sessionItem);
    const gchar* title = xbel_item_get_title(browser->sessionItem);
    GtkWidget* menuitem = menu_item_new(title, GTK_STOCK_FILE
     , G_CALLBACK(on_menu_tabsclosed_item_activate), TRUE, browser);
    sokoke_widget_set_tooltip_text(menuitem, uri);
    g_object_set_data(G_OBJECT(menuitem), "uri", (gchar*)uri);
    gtk_menu_shell_prepend(GTK_MENU_SHELL(browser->menu_tabsclosed), menuitem);
    gtk_widget_show(menuitem);
    gtk_widget_destroy(browser->webView_menu);
    xbel_folder_remove_item(session, browser->sessionItem);
    xbel_item_free(browser->sessionItem);
    gtk_notebook_remove_page(GTK_NOTEBOOK(browser->webViews)
     , get_webView_index(webView, browser));
    update_browser_actions(browser);
}

#undef webView_IS_SENSITIVE
